#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import os
from abc import ABC, abstractmethod
from datetime import timezone
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Tuple

import pandas as pd
import pendulum
from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.streams import IncrementalMixin
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer
from numpy import nan
from source_bing_ads.base_streams import Accounts, BingAdsBaseStream
from source_bing_ads.utils import transform_bulk_datetime_format_to_rfc_3339


class BingAdsBulkStream(BingAdsBaseStream, IncrementalMixin, ABC):

    transformer: TypeTransformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization | TransformConfig.CustomSchemaNormalization)
    cursor_field = "Modified Time"
    primary_key = "Id"
    _state = {}

    @staticmethod
    @transformer.registerCustomTransform
    def custom_transform_date_rfc3339(original_value, field_schema):
        if original_value and "format" in field_schema and field_schema["format"] == "date-time":
            transformed_value = transform_bulk_datetime_format_to_rfc_3339(original_value)
            return transformed_value
        return original_value

    @property
    @abstractmethod
    def data_scope(self) -> List[str]:
        """
        Defines scopes or types of data to download. Docs: https://learn.microsoft.com/en-us/advertising/bulk-service/datascope?view=bingads-13
        """

    @property
    @abstractmethod
    def download_entities(self) -> List[str]:
        """
        Defines the entities that should be downloaded. Docs: https://learn.microsoft.com/en-us/advertising/bulk-service/downloadentity?view=bingads-13
        """

    def stream_slices(
        self,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        accounts = Accounts(self.client, self.config)
        for _slice in accounts.stream_slices():
            for account in accounts.read_records(SyncMode.full_refresh, _slice):
                yield {"account_id": account["Id"], "customer_id": account["ParentCustomerId"]}

    @property
    def state(self) -> Mapping[str, Any]:
        return self._state

    @state.setter
    def state(self, value: Mapping[str, Any]):
        # if key 'Account Id' exists, so we receive a record that should be parsed to state
        # otherwise state object from connection state was received
        account_id = value.get("Account Id")

        if account_id and value[self.cursor_field]:
            current_state_value = self._state.get(str(value["Account Id"]), {}).get(self.cursor_field, "")
            record_state_value = transform_bulk_datetime_format_to_rfc_3339(value[self.cursor_field])
            new_state_value = max(current_state_value, record_state_value)
            self._state.update({str(value["Account Id"]): {self.cursor_field: new_state_value}})
        else:
            self._state.update(value)

    def get_start_date(self, stream_state: Mapping[str, Any] = None, account_id: str = None) -> Optional[pendulum.DateTime]:
        """
        The start_date in the query can only be specified if it is within a period of up to 30 days from today.
        """
        min_available_date = pendulum.now().subtract(days=30).astimezone(tz=timezone.utc)
        start_date = self.client.reports_start_date
        if stream_state.get(account_id, {}).get(self.cursor_field):
            start_date = pendulum.parse(stream_state[account_id][self.cursor_field])
        return start_date if start_date and start_date > min_available_date else None

    def read_records(
        self,
        sync_mode: SyncMode,
        stream_slice: Mapping[str, Any] = None,
        stream_state: Mapping[str, Any] = None,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Mapping[str, Any]]:
        stream_state = stream_state or {}
        account_id = str(stream_slice.get("account_id")) if stream_slice else None
        customer_id = str(stream_slice.get("customer_id")) if stream_slice else None

        report_file_path = self.client.get_bulk_entity(
            data_scope=self.data_scope,
            download_entities=self.download_entities,
            customer_id=customer_id,
            account_id=account_id,
            start_date=self.get_start_date(stream_state, account_id),
        )
        for record in self.read_with_chunks(report_file_path):
            record = self.transform(record, stream_slice)
            yield record
            self.state = record

    def read_with_chunks(self, path: str, chunk_size: int = 1024) -> Iterable[Tuple[int, Mapping[str, Any]]]:
        try:
            with open(path, "r") as data:
                chunks = pd.read_csv(data, chunksize=chunk_size, iterator=True, dialect="unix", dtype=object)
                for chunk in chunks:
                    chunk = chunk.replace({nan: None}).to_dict(orient="records")
                    for row in chunk:
                        if row.get("Type") not in ("Format Version", "Account"):
                            yield row
        except pd.errors.EmptyDataError as e:
            self.logger.info(f"Empty data received. {e}")
        except IOError as ioe:
            self.logger.fatal(
                f"The IO/Error occurred while reading tmp data. Called: {path}. Stream: {self.name}",
            )
            raise ioe
        finally:
            # remove binary tmp file, after data is read
            os.remove(path)

    def transform(self, record: MutableMapping[str, Any], stream_slice: Mapping[str, Any], **kwargs) -> MutableMapping[str, Any]:
        """
        Bing Ads Bulk API returns all available properties for all entities.
        This method filter out only available properties.
        """
        actual_record = {key: value for key, value in record.items() if key in self.get_json_schema()["properties"].keys()}
        actual_record["Account Id"] = stream_slice.get("account_id")
        return actual_record


class AppInstallAds(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/app-install-ad?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["AppInstallAds"]


class AppInstallAdLabels(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/app-install-ad-label?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["AppInstallAdLabels"]


class Labels(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/label?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["Labels"]


class KeywordLabels(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/keyword-label?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["KeywordLabels"]


class Keywords(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/keyword?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["Keywords"]


class CampaignLabels(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/campaign-label?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["CampaignLabels"]


class AdGroupLabels(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/ad-group-label?view=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["AdGroupLabels"]


class Budget(BingAdsBulkStream):
    """
    https://learn.microsoft.com/en-us/advertising/bulk-service/budget?view=bingads-13&viewFallbackFrom=bingads-13
    """

    data_scope = ["EntityData"]
    download_entities = ["Budgets"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
from itertools import product
from typing import Any, List, Mapping, Optional, Tuple

from airbyte_cdk.models import FailureType, SyncMode
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.utils import AirbyteTracedException
from source_bing_ads.base_streams import Accounts, AdGroups, Ads, Campaigns
from source_bing_ads.bulk_streams import (
    AdGroupLabels,
    AppInstallAdLabels,
    AppInstallAds,
    Budget,
    CampaignLabels,
    KeywordLabels,
    Keywords,
    Labels,
)
from source_bing_ads.client import Client
from source_bing_ads.report_streams import (  # noqa: F401
    AccountImpressionPerformanceReportDaily,
    AccountImpressionPerformanceReportHourly,
    AccountImpressionPerformanceReportMonthly,
    AccountImpressionPerformanceReportWeekly,
    AccountPerformanceReportDaily,
    AccountPerformanceReportHourly,
    AccountPerformanceReportMonthly,
    AccountPerformanceReportWeekly,
    AdGroupImpressionPerformanceReportDaily,
    AdGroupImpressionPerformanceReportHourly,
    AdGroupImpressionPerformanceReportMonthly,
    AdGroupImpressionPerformanceReportWeekly,
    AdGroupPerformanceReportDaily,
    AdGroupPerformanceReportHourly,
    AdGroupPerformanceReportMonthly,
    AdGroupPerformanceReportWeekly,
    AdPerformanceReportDaily,
    AdPerformanceReportHourly,
    AdPerformanceReportMonthly,
    AdPerformanceReportWeekly,
    AgeGenderAudienceReportDaily,
    AgeGenderAudienceReportHourly,
    AgeGenderAudienceReportMonthly,
    AgeGenderAudienceReportWeekly,
    AudiencePerformanceReportDaily,
    AudiencePerformanceReportHourly,
    AudiencePerformanceReportMonthly,
    AudiencePerformanceReportWeekly,
    BingAdsReportingServiceStream,
    BudgetSummaryReport,
    CampaignImpressionPerformanceReportDaily,
    CampaignImpressionPerformanceReportHourly,
    CampaignImpressionPerformanceReportMonthly,
    CampaignImpressionPerformanceReportWeekly,
    CampaignPerformanceReportDaily,
    CampaignPerformanceReportHourly,
    CampaignPerformanceReportMonthly,
    CampaignPerformanceReportWeekly,
    CustomReport,
    GeographicPerformanceReportDaily,
    GeographicPerformanceReportHourly,
    GeographicPerformanceReportMonthly,
    GeographicPerformanceReportWeekly,
    GoalsAndFunnelsReportDaily,
    GoalsAndFunnelsReportHourly,
    GoalsAndFunnelsReportMonthly,
    GoalsAndFunnelsReportWeekly,
    KeywordPerformanceReportDaily,
    KeywordPerformanceReportHourly,
    KeywordPerformanceReportMonthly,
    KeywordPerformanceReportWeekly,
    ProductDimensionPerformanceReportDaily,
    ProductDimensionPerformanceReportHourly,
    ProductDimensionPerformanceReportMonthly,
    ProductDimensionPerformanceReportWeekly,
    ProductSearchQueryPerformanceReportDaily,
    ProductSearchQueryPerformanceReportHourly,
    ProductSearchQueryPerformanceReportMonthly,
    ProductSearchQueryPerformanceReportWeekly,
    SearchQueryPerformanceReportDaily,
    SearchQueryPerformanceReportHourly,
    SearchQueryPerformanceReportMonthly,
    SearchQueryPerformanceReportWeekly,
    UserLocationPerformanceReportDaily,
    UserLocationPerformanceReportHourly,
    UserLocationPerformanceReportMonthly,
    UserLocationPerformanceReportWeekly,
)


class SourceBingAds(AbstractSource):
    """
    Source implementation of Bing Ads API. Fetches advertising data from accounts
    """

    def check_connection(self, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, Any]:
        try:
            client = Client(**config)
            accounts = Accounts(client, config)
            account_ids = set()
            for _slice in accounts.stream_slices():
                account_ids.update({str(account["Id"]) for account in accounts.read_records(SyncMode.full_refresh, _slice)})
            self.validate_custom_reposts(config, client)
            if account_ids:
                return True, None
            else:
                raise AirbyteTracedException(
                    message="Config validation error: You don't have accounts assigned to this user. Please verify your developer token.",
                    internal_message="You don't have accounts assigned to this user.",
                    failure_type=FailureType.config_error,
                )
        except Exception as error:
            return False, error

    def validate_custom_reposts(self, config: Mapping[str, Any], client: Client):
        custom_reports = self.get_custom_reports(config, client)
        for custom_report in custom_reports:
            is_valid, reason = custom_report.validate_report_configuration()
            if not is_valid:
                raise AirbyteTracedException(
                    message=f"Config validation error: {custom_report.name}: {reason}",
                    internal_message=f"{custom_report.name}: {reason}",
                    failure_type=FailureType.config_error,
                )

    def _clear_reporting_object_name(self, report_object: str) -> str:
        # reporting mixin adds it
        if report_object.endswith("Request"):
            return report_object.replace("Request", "")
        return report_object

    def get_custom_reports(self, config: Mapping[str, Any], client: Client) -> List[Optional[Stream]]:
        return [
            type(
                report["name"],
                (CustomReport,),
                {
                    "report_name": self._clear_reporting_object_name(report["reporting_object"]),
                    "custom_report_columns": report["report_columns"],
                    "report_aggregation": report["report_aggregation"],
                },
            )(client, config)
            for report in config.get("custom_reports", [])
        ]

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        client = Client(**config)
        streams = [
            Accounts(client, config),
            AdGroups(client, config),
            AdGroupLabels(client, config),
            AppInstallAds(client, config),
            AppInstallAdLabels(client, config),
            Ads(client, config),
            Budget(client, config),
            Campaigns(client, config),
            BudgetSummaryReport(client, config),
            Labels(client, config),
            KeywordLabels(client, config),
            Keywords(client, config),
            CampaignLabels(client, config),
        ]

        reports = (
            "AgeGenderAudienceReport",
            "AccountImpressionPerformanceReport",
            "AccountPerformanceReport",
            "AudiencePerformanceReport",
            "KeywordPerformanceReport",
            "AdGroupPerformanceReport",
            "AdPerformanceReport",
            "AdGroupImpressionPerformanceReport",
            "CampaignPerformanceReport",
            "CampaignImpressionPerformanceReport",
            "GeographicPerformanceReport",
            "GoalsAndFunnelsReport",
            "ProductDimensionPerformanceReport",
            "ProductSearchQueryPerformanceReport",
            "SearchQueryPerformanceReport",
            "UserLocationPerformanceReport",
        )
        report_aggregation = ("Hourly", "Daily", "Weekly", "Monthly")
        streams.extend([eval(f"{report}{aggregation}")(client, config) for (report, aggregation) in product(reports, report_aggregation)])

        custom_reports = self.get_custom_reports(config, client)
        streams.extend(custom_reports)
        return streams



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import re
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Set, Tuple, Union
from urllib.parse import urlparse

import _csv
import pendulum
from airbyte_cdk.sources.streams.core import package_name_from_class
from airbyte_cdk.sources.utils.schema_helpers import ResourceSchemaLoader
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer
from airbyte_protocol.models import SyncMode
from bingads import ServiceClient
from bingads.v13.internal.reporting.row_report import _RowReport
from bingads.v13.internal.reporting.row_report_iterator import _RowReportRecord
from bingads.v13.reporting import ReportingDownloadParameters
from cached_property import cached_property
from source_bing_ads.base_streams import Accounts, BingAdsStream
from source_bing_ads.utils import transform_date_format_to_rfc_3339, transform_report_hourly_datetime_format_to_rfc_3339
from suds import WebFault, sudsobject


class HourlyReportTransformerMixin:
    transformer: TypeTransformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization | TransformConfig.CustomSchemaNormalization)

    @staticmethod
    @transformer.registerCustomTransform
    def custom_transform_datetime_rfc3339(original_value, field_schema):
        if original_value and "format" in field_schema and field_schema["format"] == "date-time":
            transformed_value = transform_report_hourly_datetime_format_to_rfc_3339(original_value)
            return transformed_value
        return original_value


class BingAdsReportingServiceStream(BingAdsStream, ABC):
    # The directory where the file with report will be downloaded.
    file_directory: str = "/tmp"
    # timeout for reporting download operations in milliseconds
    timeout: int = 300000
    report_file_format: str = "Csv"

    transformer: TypeTransformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization)
    primary_key: List[str] = ["TimePeriod", "Network", "DeviceType"]

    cursor_field = "TimePeriod"
    service_name: str = "ReportingService"
    operation_name: str = "download_report"

    def get_json_schema(self) -> Mapping[str, Any]:
        return ResourceSchemaLoader(package_name_from_class(self.__class__)).get_schema(self.report_schema_name)

    @property
    @abstractmethod
    def report_name(self) -> str:
        """
        Specifies bing ads report naming
        """

    @property
    @abstractmethod
    def report_aggregation(self) -> Optional[str]:
        """
        Specifies bing ads report aggregation type
        Supported types: Hourly, Daily, Weekly, Monthly
        """

    @property
    @abstractmethod
    def report_schema_name(self) -> str:
        """
        Specifies file name with schema
        """

    @property
    def default_time_periods(self):
        # used when reports start date is not provided
        return ["LastYear", "ThisYear"] if self.report_aggregation not in ("DayOfWeek", "HourOfDay") else ["ThisYear"]

    @property
    def report_columns(self) -> Iterable[str]:
        return list(self.get_json_schema().get("properties", {}).keys())

    def parse_response(self, response: sudsobject.Object, **kwargs: Mapping[str, Any]) -> Iterable[Mapping]:
        if response is not None:
            try:
                for row in response.report_records:
                    yield {column: self.get_column_value(row, column) for column in self.report_columns}
            except _csv.Error as e:
                self.logger.warning(f"CSV report file for stream `{self.name}` is broken or cannot be read correctly: {e}, skipping ...")

    def get_column_value(self, row: _RowReportRecord, column: str) -> Union[str, None, int, float]:
        """
        Reads field value from row and transforms:
        1. empty values to logical None
        2. Percent values to numeric string e.g. "12.25%" -> "12.25"
        """
        value = row.value(column)
        if not value or value == "--":
            return None
        if "%" in value:
            value = value.replace("%", "")
        if value and column in self._get_schema_numeric_properties:
            value = value.replace(",", "")
        return value

    @cached_property
    def _get_schema_numeric_properties(self) -> Set[str]:
        return set(k for k, v in self.get_json_schema()["properties"].items() if set(v.get("type")) & {"integer", "number"})

    def get_request_date(self, reporting_service: ServiceClient, date: datetime) -> sudsobject.Object:
        """
        Creates XML Date object based on datetime.
        https://docs.microsoft.com/en-us/advertising/reporting-service/date?view=bingads-13
        The [suds.client.Factory-class.html factory] namespace provides a factory that may be used
        to create instances of objects and types defined in the WSDL.
        """
        request_date = reporting_service.factory.create("Date")
        request_date.Day = date.day
        request_date.Month = date.month
        request_date.Year = date.year
        return request_date

    def request_params(
        self, stream_state: Mapping[str, Any] = None, account_id: str = None, **kwargs: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        stream_slice = kwargs["stream_slice"]
        start_date = self.get_start_date(stream_state, account_id)

        reporting_service = self.client.get_service("ReportingService")
        request_time_zone = reporting_service.factory.create("ReportTimeZone")

        report_time = reporting_service.factory.create("ReportTime")
        report_time.ReportTimeZone = request_time_zone.GreenwichMeanTimeDublinEdinburghLisbonLondon
        if start_date:
            report_time.CustomDateRangeStart = self.get_request_date(reporting_service, start_date)
            report_time.CustomDateRangeEnd = self.get_request_date(reporting_service, datetime.utcnow())
            report_time.PredefinedTime = None
        else:
            report_time.CustomDateRangeStart = None
            report_time.CustomDateRangeEnd = None
            report_time.PredefinedTime = stream_slice["time_period"]

        report_request = self.get_report_request(account_id, False, False, False, self.report_file_format, False, report_time)

        return {
            "report_request": report_request,
            "result_file_directory": self.file_directory,
            "result_file_name": self.report_name,
            "overwrite_result_file": True,
            "timeout_in_milliseconds": self.timeout,
        }

    def get_start_date(self, stream_state: Mapping[str, Any] = None, account_id: str = None):
        if stream_state and account_id:
            if stream_state.get(account_id, {}).get(self.cursor_field):
                return pendulum.parse(stream_state[account_id][self.cursor_field])

        return self.client.reports_start_date

    def get_updated_state(
        self,
        current_stream_state: MutableMapping[str, Any],
        latest_record: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        account_id = str(latest_record["AccountId"])
        current_stream_state[account_id] = current_stream_state.get(account_id, {})
        current_stream_state[account_id][self.cursor_field] = max(
            self.get_report_record_timestamp(latest_record[self.cursor_field]),
            current_stream_state.get(account_id, {}).get(self.cursor_field, ""),
        )
        return current_stream_state

    def send_request(self, params: Mapping[str, Any], customer_id: str, account_id: str) -> _RowReport:
        request_kwargs = {
            "service_name": None,
            "customer_id": customer_id,
            "account_id": account_id,
            "operation_name": self.operation_name,
            "is_report_service": True,
            "params": {"download_parameters": ReportingDownloadParameters(**params)},
        }
        return self.client.request(**request_kwargs)

    def get_report_request(
        self,
        account_id: str,
        exclude_column_headers: bool,
        exclude_report_footer: bool,
        exclude_report_header: bool,
        report_file_format: str,
        return_only_complete_data: bool,
        time: sudsobject.Object,
    ) -> sudsobject.Object:
        reporting_service = self.client.get_service(self.service_name)
        report_request = reporting_service.factory.create(f"{self.report_name}Request")
        if self.report_aggregation:
            report_request.Aggregation = self.report_aggregation

        report_request.ExcludeColumnHeaders = exclude_column_headers
        report_request.ExcludeReportFooter = exclude_report_footer
        report_request.ExcludeReportHeader = exclude_report_header
        report_request.Format = report_file_format
        report_request.FormatVersion = "2.0"
        report_request.ReturnOnlyCompleteData = return_only_complete_data
        report_request.Time = time
        report_request.ReportName = self.report_name
        # Defines the set of accounts and campaigns to include in the report.
        scope = reporting_service.factory.create("AccountThroughCampaignReportScope")
        scope.AccountIds = {"long": [account_id]}
        scope.Campaigns = None
        report_request.Scope = scope

        columns = reporting_service.factory.create(f"ArrayOf{self.report_name}Column")
        getattr(columns, f"{self.report_name}Column").append(self.report_columns)
        report_request.Columns = columns
        return report_request

    def get_report_record_timestamp(self, datestring: str) -> str:
        """
        Parse report date field based on aggregation type
        """
        return (
            self.transformer._custom_normalizer(datestring, self.get_json_schema()["properties"][self.cursor_field])
            if self.transformer._custom_normalizer
            else datestring
        )

    def stream_slices(
        self, *, sync_mode: SyncMode, cursor_field: Optional[List[str]] = None, stream_state: Optional[Mapping[str, Any]] = None
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        accounts = Accounts(self.client, self.config)
        for _slice in accounts.stream_slices():
            for account in accounts.read_records(SyncMode.full_refresh, _slice):
                if self.get_start_date(stream_state, account["Id"]):  # if start date is not provided default time periods will be used
                    yield {"account_id": account["Id"], "customer_id": account["ParentCustomerId"]}
                else:
                    for period in self.default_time_periods:
                        yield {"account_id": account["Id"], "customer_id": account["ParentCustomerId"], "time_period": period}


class BingAdsReportingServicePerformanceStream(BingAdsReportingServiceStream, ABC):
    def get_start_date(self, stream_state: Mapping[str, Any] = None, account_id: str = None):
        start_date = super().get_start_date(stream_state, account_id)

        if self.config.get("lookback_window") and start_date:
            # Datetime subtract won't work with days = 0
            # it'll output an AirbyteError
            return start_date.subtract(days=self.config["lookback_window"])
        else:
            return start_date


class BudgetSummaryReport(BingAdsReportingServiceStream):
    report_name: str = "BudgetSummaryReport"
    report_aggregation = None
    cursor_field = "Date"
    report_schema_name = "budget_summary_report"
    primary_key = "Date"

    transformer: TypeTransformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization | TransformConfig.CustomSchemaNormalization)

    @staticmethod
    @transformer.registerCustomTransform
    def custom_transform_date_rfc3339(original_value, field_schema):
        if original_value and "format" in field_schema and field_schema["format"] == "date":
            transformed_value = transform_date_format_to_rfc_3339(original_value)
            return transformed_value
        return original_value


class CampaignPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "CampaignPerformanceReport"

    report_schema_name = "campaign_performance_report"
    primary_key = [
        "AccountId",
        "CampaignId",
        "TimePeriod",
        "CurrencyCode",
        "AdDistribution",
        "DeviceType",
        "Network",
        "DeliveredMatchType",
        "DeviceOS",
        "TopVsOther",
        "BidMatchType",
    ]


class CampaignPerformanceReportHourly(HourlyReportTransformerMixin, CampaignPerformanceReport):
    report_aggregation = "Hourly"

    report_schema_name = "campaign_performance_report_hourly"


class CampaignPerformanceReportDaily(CampaignPerformanceReport):
    report_aggregation = "Daily"


class CampaignPerformanceReportWeekly(CampaignPerformanceReport):
    report_aggregation = "Weekly"


class CampaignPerformanceReportMonthly(CampaignPerformanceReport):
    report_aggregation = "Monthly"


class CampaignImpressionPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    https://learn.microsoft.com/en-us/advertising/reporting-service/adgroupperformancereportrequest?view=bingads-13
    Primary key cannot be set: due to included `Impression Share Performance Statistics` some fields should be removed,
    see https://learn.microsoft.com/en-us/advertising/guides/reports?view=bingads-13#columnrestrictions for more info.
    """

    report_name: str = "CampaignPerformanceReport"

    report_schema_name = "campaign_impression_performance_report"

    primary_key = None


class CampaignImpressionPerformanceReportHourly(HourlyReportTransformerMixin, CampaignImpressionPerformanceReport):
    report_aggregation = "Hourly"

    report_schema_name = "campaign_impression_performance_report_hourly"


class CampaignImpressionPerformanceReportDaily(CampaignImpressionPerformanceReport):
    report_aggregation = "Daily"


class CampaignImpressionPerformanceReportWeekly(CampaignImpressionPerformanceReport):
    report_aggregation = "Weekly"


class CampaignImpressionPerformanceReportMonthly(CampaignImpressionPerformanceReport):
    report_aggregation = "Monthly"


class AdPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "AdPerformanceReport"

    report_schema_name = "ad_performance_report"
    primary_key = [
        "AccountId",
        "CampaignId",
        "AdGroupId",
        "AdId",
        "TimePeriod",
        "CurrencyCode",
        "AdDistribution",
        "DeviceType",
        "Language",
        "Network",
        "DeviceOS",
        "TopVsOther",
        "BidMatchType",
        "DeliveredMatchType",
    ]


class AdPerformanceReportHourly(HourlyReportTransformerMixin, AdPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "ad_performance_report_hourly"


class AdPerformanceReportDaily(AdPerformanceReport):
    report_aggregation = "Daily"


class AdPerformanceReportWeekly(AdPerformanceReport):
    report_aggregation = "Weekly"


class AdPerformanceReportMonthly(AdPerformanceReport):
    report_aggregation = "Monthly"


class AdGroupPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "AdGroupPerformanceReport"
    report_schema_name = "ad_group_performance_report"

    primary_key = [
        "AccountId",
        "CampaignId",
        "AdGroupId",
        "TimePeriod",
        "CurrencyCode",
        "AdDistribution",
        "DeviceType",
        "Network",
        "DeliveredMatchType",
        "DeviceOS",
        "TopVsOther",
        "BidMatchType",
        "Language",
    ]


class AdGroupPerformanceReportHourly(HourlyReportTransformerMixin, AdGroupPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "ad_group_performance_report_hourly"


class AdGroupPerformanceReportDaily(AdGroupPerformanceReport):
    report_aggregation = "Daily"


class AdGroupPerformanceReportWeekly(AdGroupPerformanceReport):
    report_aggregation = "Weekly"


class AdGroupPerformanceReportMonthly(AdGroupPerformanceReport):
    report_aggregation = "Monthly"


class AdGroupImpressionPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    https://learn.microsoft.com/en-us/advertising/reporting-service/adgroupperformancereportrequest?view=bingads-13
    Primary key cannot be set: due to included `Impression Share Performance Statistics` some fields should be removed,
    see https://learn.microsoft.com/en-us/advertising/guides/reports?view=bingads-13#columnrestrictions for more info.
    """

    report_name: str = "AdGroupPerformanceReport"
    report_schema_name = "ad_group_impression_performance_report"


class AdGroupImpressionPerformanceReportHourly(HourlyReportTransformerMixin, AdGroupImpressionPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "ad_group_impression_performance_report_hourly"


class AdGroupImpressionPerformanceReportDaily(AdGroupImpressionPerformanceReport):
    report_aggregation = "Daily"


class AdGroupImpressionPerformanceReportWeekly(AdGroupImpressionPerformanceReport):
    report_aggregation = "Weekly"


class AdGroupImpressionPerformanceReportMonthly(AdGroupImpressionPerformanceReport):
    report_aggregation = "Monthly"


class KeywordPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "KeywordPerformanceReport"
    report_schema_name = "keyword_performance_report"
    primary_key = [
        "AccountId",
        "CampaignId",
        "AdGroupId",
        "KeywordId",
        "AdId",
        "TimePeriod",
        "CurrencyCode",
        "DeliveredMatchType",
        "AdDistribution",
        "DeviceType",
        "Language",
        "Network",
        "DeviceOS",
        "TopVsOther",
        "BidMatchType",
    ]


class KeywordPerformanceReportHourly(HourlyReportTransformerMixin, KeywordPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "keyword_performance_report_hourly"


class KeywordPerformanceReportDaily(KeywordPerformanceReport):
    report_aggregation = "Daily"
    report_schema_name = "keyword_performance_report_daily"


class KeywordPerformanceReportWeekly(KeywordPerformanceReport):
    report_aggregation = "Weekly"


class KeywordPerformanceReportMonthly(KeywordPerformanceReport):
    report_aggregation = "Monthly"


class GeographicPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "GeographicPerformanceReport"
    report_schema_name = "geographic_performance_report"

    # Need to override the primary key here because the one inherited from the PerformanceReportsMixin
    # is incorrect for the geographic performance reports
    primary_key = None


class GeographicPerformanceReportHourly(HourlyReportTransformerMixin, GeographicPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "geographic_performance_report_hourly"


class GeographicPerformanceReportDaily(GeographicPerformanceReport):
    report_aggregation = "Daily"


class GeographicPerformanceReportWeekly(GeographicPerformanceReport):
    report_aggregation = "Weekly"


class GeographicPerformanceReportMonthly(GeographicPerformanceReport):
    report_aggregation = "Monthly"


class AccountPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "AccountPerformanceReport"
    report_schema_name = "account_performance_report"
    primary_key = [
        "AccountId",
        "TimePeriod",
        "CurrencyCode",
        "AdDistribution",
        "DeviceType",
        "Network",
        "DeliveredMatchType",
        "DeviceOS",
        "TopVsOther",
        "BidMatchType",
    ]


class AccountPerformanceReportHourly(HourlyReportTransformerMixin, AccountPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "account_performance_report_hourly"


class AccountPerformanceReportDaily(AccountPerformanceReport):
    report_aggregation = "Daily"


class AccountPerformanceReportWeekly(AccountPerformanceReport):
    report_aggregation = "Weekly"


class AccountPerformanceReportMonthly(AccountPerformanceReport):
    report_aggregation = "Monthly"


class AccountImpressionPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    Report source: https://docs.microsoft.com/en-us/advertising/reporting-service/accountperformancereportrequest?view=bingads-13
    Primary key cannot be set: due to included `Impression Share Performance Statistics` some fields should be removed,
    see https://learn.microsoft.com/en-us/advertising/guides/reports?view=bingads-13#columnrestrictions for more info.
    """

    report_name: str = "AccountPerformanceReport"
    report_schema_name = "account_impression_performance_report"
    primary_key = None


class AccountImpressionPerformanceReportHourly(HourlyReportTransformerMixin, AccountImpressionPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "account_impression_performance_report_hourly"


class AccountImpressionPerformanceReportDaily(AccountImpressionPerformanceReport):
    report_aggregation = "Daily"


class AccountImpressionPerformanceReportWeekly(AccountImpressionPerformanceReport):
    report_aggregation = "Weekly"


class AccountImpressionPerformanceReportMonthly(AccountImpressionPerformanceReport):
    report_aggregation = "Monthly"


class AgeGenderAudienceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "AgeGenderAudienceReport"

    report_schema_name = "age_gender_audience_report"
    primary_key = ["AgeGroup", "Gender", "TimePeriod", "AccountId", "CampaignId", "Language", "AdDistribution"]


class AgeGenderAudienceReportHourly(HourlyReportTransformerMixin, AgeGenderAudienceReport):
    report_aggregation = "Hourly"
    report_schema_name = "age_gender_audience_report_hourly"


class AgeGenderAudienceReportDaily(AgeGenderAudienceReport):
    report_aggregation = "Daily"


class AgeGenderAudienceReportWeekly(AgeGenderAudienceReport):
    report_aggregation = "Weekly"


class AgeGenderAudienceReportMonthly(AgeGenderAudienceReport):
    report_aggregation = "Monthly"


class SearchQueryPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "SearchQueryPerformanceReport"
    report_schema_name = "search_query_performance_report"

    primary_key = [
        "SearchQuery",
        "Keyword",
        "TimePeriod",
        "AccountId",
        "CampaignId",
        "Language",
        "DeliveredMatchType",
        "DeviceType",
        "DeviceOS",
        "TopVsOther",
    ]


class SearchQueryPerformanceReportHourly(HourlyReportTransformerMixin, SearchQueryPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "search_query_performance_report_hourly"


class SearchQueryPerformanceReportDaily(SearchQueryPerformanceReport):
    report_aggregation = "Daily"


class SearchQueryPerformanceReportWeekly(SearchQueryPerformanceReport):
    report_aggregation = "Weekly"


class SearchQueryPerformanceReportMonthly(SearchQueryPerformanceReport):
    report_aggregation = "Monthly"


class UserLocationPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    report_name: str = "UserLocationPerformanceReport"
    report_schema_name = "user_location_performance_report"
    primary_key = [
        "AccountId",
        "AdGroupId",
        "CampaignId",
        "DeliveredMatchType",
        "DeviceOS",
        "DeviceType",
        "Language",
        "LocationId",
        "QueryIntentLocationId",
        "TimePeriod",
        "TopVsOther",
    ]


class UserLocationPerformanceReportHourly(HourlyReportTransformerMixin, UserLocationPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "user_location_performance_report_hourly"


class UserLocationPerformanceReportDaily(UserLocationPerformanceReport):
    report_aggregation = "Daily"


class UserLocationPerformanceReportWeekly(UserLocationPerformanceReport):
    report_aggregation = "Weekly"


class UserLocationPerformanceReportMonthly(UserLocationPerformanceReport):
    report_aggregation = "Monthly"


class ProductDimensionPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    https://learn.microsoft.com/en-us/advertising/reporting-service/productdimensionperformancereportrequest?view=bingads-13
    """

    report_name: str = "ProductDimensionPerformanceReport"
    report_schema_name = "product_dimension_performance_report"
    primary_key = None

    @property
    def report_columns(self) -> Iterable[str]:
        """AccountId is not in reporting columns for this report"""
        properties = list(self.get_json_schema().get("properties", {}).keys())
        properties.remove("AccountId")
        return properties

    def transform(self, record: MutableMapping[str, Any], stream_slice: Mapping[str, Any], **kwargs) -> MutableMapping[str, Any]:
        record = super().transform(record, stream_slice)
        record["AccountId"] = stream_slice["account_id"]
        return record


class ProductDimensionPerformanceReportHourly(HourlyReportTransformerMixin, ProductDimensionPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "product_dimension_performance_report_hourly"


class ProductDimensionPerformanceReportDaily(ProductDimensionPerformanceReport):
    report_aggregation = "Daily"


class ProductDimensionPerformanceReportWeekly(ProductDimensionPerformanceReport):
    report_aggregation = "Weekly"


class ProductDimensionPerformanceReportMonthly(ProductDimensionPerformanceReport):
    report_aggregation = "Monthly"


class ProductSearchQueryPerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    https://learn.microsoft.com/en-us/advertising/reporting-service/productsearchqueryperformancereportrequest?view=bingads-13
    """

    report_name: str = "ProductSearchQueryPerformanceReport"
    report_schema_name = "product_search_query_performance_report"
    primary_key = [
        "AccountId",
        "TimePeriod",
        "CampaignId",
        "AdId",
        "AdGroupId",
        "SearchQuery",
        "DeviceType",
        "DeviceOS",
        "Language",
        "Network",
    ]


class ProductSearchQueryPerformanceReportHourly(HourlyReportTransformerMixin, ProductSearchQueryPerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "product_search_query_performance_report_hourly"


class ProductSearchQueryPerformanceReportDaily(ProductSearchQueryPerformanceReport):
    report_aggregation = "Daily"


class ProductSearchQueryPerformanceReportWeekly(ProductSearchQueryPerformanceReport):
    report_aggregation = "Weekly"


class ProductSearchQueryPerformanceReportMonthly(ProductSearchQueryPerformanceReport):
    report_aggregation = "Monthly"


class GoalsAndFunnelsReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    https://learn.microsoft.com/en-us/advertising/reporting-service/goalsandfunnelsreportrequest?view=bingads-13
    """

    report_name: str = "GoalsAndFunnelsReport"
    report_schema_name = "goals_and_funnels_report"
    primary_key = [
        "GoalId",
        "TimePeriod",
        "AccountId",
        "CampaignId",
        "DeviceType",
        "DeviceOS",
        "AdGroupId",
    ]


class GoalsAndFunnelsReportHourly(HourlyReportTransformerMixin, GoalsAndFunnelsReport):
    report_aggregation = "Hourly"
    report_schema_name = "goals_and_funnels_report_hourly"


class GoalsAndFunnelsReportDaily(GoalsAndFunnelsReport):
    report_aggregation = "Daily"


class GoalsAndFunnelsReportWeekly(GoalsAndFunnelsReport):
    report_aggregation = "Weekly"


class GoalsAndFunnelsReportMonthly(GoalsAndFunnelsReport):
    report_aggregation = "Monthly"


class AudiencePerformanceReport(BingAdsReportingServicePerformanceStream, ABC):
    """
    https://learn.microsoft.com/en-us/advertising/reporting-service/audienceperformancereportrequest?view=bingads-13
    """

    report_name: str = "AudiencePerformanceReport"
    report_schema_name = "audience_performance_report"
    primary_key = [
        "AudienceId",
        "TimePeriod",
        "AccountId",
        "CampaignId",
        "AdGroupId",
    ]


class AudiencePerformanceReportHourly(HourlyReportTransformerMixin, AudiencePerformanceReport):
    report_aggregation = "Hourly"
    report_schema_name = "audience_performance_report_hourly"


class AudiencePerformanceReportDaily(AudiencePerformanceReport):
    report_aggregation = "Daily"


class AudiencePerformanceReportWeekly(AudiencePerformanceReport):
    report_aggregation = "Weekly"


class AudiencePerformanceReportMonthly(AudiencePerformanceReport):
    report_aggregation = "Monthly"


class CustomReport(BingAdsReportingServicePerformanceStream, ABC):
    transformer: TypeTransformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization)
    custom_report_columns = []
    report_schema_name = None
    primary_key = None

    @property
    def cursor_field(self) -> Union[str, List[str]]:
        # Summary aggregation doesn't include TimePeriod field
        if self.report_aggregation not in ("Summary", "DayOfWeek", "HourOfDay"):
            return "TimePeriod"

    @property
    def report_columns(self):
        # adding common and default columns
        if "AccountId" not in self.custom_report_columns:
            self.custom_report_columns.append("AccountId")
        if self.cursor_field and self.cursor_field not in self.custom_report_columns:
            self.custom_report_columns.append(self.cursor_field)
        return list(frozenset(self.custom_report_columns))

    def get_json_schema(self) -> Mapping[str, Any]:
        columns_schema = {col: {"type": ["null", "string"]} for col in self.report_columns}
        schema: Mapping[str, Any] = {
            "$schema": "https://json-schema.org/draft-07/schema#",
            "type": ["null", "object"],
            "additionalProperties": True,
            "properties": columns_schema,
        }
        return schema

    def validate_report_configuration(self) -> Tuple[bool, str]:
        # gets /bingads/v13/proxies/production/reporting_service.xml
        reporting_service_file = self.client.get_service(self.service_name)._get_service_info_dict(self.client.api_version)[
            ("reporting", self.client.environment)
        ]
        tree = ET.parse(urlparse(reporting_service_file).path)
        request_object = tree.find(f".//{{*}}complexType[@name='{self.report_name}Request']")

        report_object_columns = self._get_object_columns(request_object, tree)
        is_custom_cols_in_report_object_cols = all(x in report_object_columns for x in self.custom_report_columns)

        if not is_custom_cols_in_report_object_cols:
            return False, (
                f"Reporting Columns are invalid. Columns that you provided don't belong to Reporting Data Object Columns:"
                f" {self.custom_report_columns}. Please ensure it is correct in Bing Ads Docs."
            )

        return True, ""

    def _clear_namespace(self, type: str) -> str:
        return re.sub(r"^[a-z]+:", "", type)

    def _get_object_columns(self, request_el: ET.Element, tree: ET.ElementTree) -> List[str]:
        column_el = request_el.find(".//{*}element[@name='Columns']")
        array_of_columns_name = self._clear_namespace(column_el.get("type"))

        array_of_columns_elements = tree.find(f".//{{*}}complexType[@name='{array_of_columns_name}']")
        inner_array_of_columns_elements = array_of_columns_elements.find(".//{*}element")
        column_el_name = self._clear_namespace(inner_array_of_columns_elements.get("type"))

        column_el = tree.find(f".//{{*}}simpleType[@name='{column_el_name}']")
        column_enum_items = column_el.findall(".//{*}enumeration")
        column_enum_items_values = [el.get("value") for el in column_enum_items]
        return column_enum_items_values

    def get_report_record_timestamp(self, datestring: str) -> str:
        """
        Parse report date field based on aggregation type
        """
        if not self.report_aggregation or self.report_aggregation == "Summary":
            datestring = transform_date_format_to_rfc_3339(datestring)
        elif self.report_aggregation == "Hourly":
            datestring = transform_report_hourly_datetime_format_to_rfc_3339(datestring)
        return datestring

    def send_request(self, params: Mapping[str, Any], customer_id: str, account_id: str) -> _RowReport:
        try:
            return super().send_request(params, customer_id, account_id)
        except WebFault as e:
            self.logger.error(
                f"Could not sync custom report {self.name}: Please validate your column and aggregation configuration. "
                f"Error form server: [{e.fault.faultstring}]"
            )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import ssl
import time
from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Union
from urllib.error import URLError

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.streams import Stream
from bingads.service_client import ServiceClient
from bingads.v13.reporting.reporting_service_manager import ReportingServiceManager
from source_bing_ads.client import Client
from suds import sudsobject


class BingAdsBaseStream(Stream, ABC):
    primary_key: Optional[Union[str, List[str], List[List[str]]]] = None

    def __init__(self, client: Client, config: Mapping[str, Any]) -> None:
        super().__init__()
        self.client = client
        self.config = config


class BingAdsStream(BingAdsBaseStream, ABC):
    @property
    @abstractmethod
    def operation_name(self) -> str:
        """
        Specifies operation name to use for a current stream
        """

    @property
    @abstractmethod
    def service_name(self) -> str:
        """
        Specifies bing ads service name for a current stream
        """

    @property
    def parent_key_to_foreign_key_map(self) -> MutableMapping[str, str]:
        """
        Specifies dict with field in record as kay and slice key as value to be inserted in record in transform method.
        """
        return {}

    def transform(self, record: MutableMapping[str, Any], stream_slice: Mapping[str, Any], **kwargs) -> MutableMapping[str, Any]:
        foreign_keys = {key: stream_slice.get(value) for key, value in self.parent_key_to_foreign_key_map.items()}
        return record | foreign_keys

    @property
    def _service(self) -> Union[ServiceClient, ReportingServiceManager]:
        return self.client.get_service(service_name=self.service_name)

    @property
    def _user_id(self) -> int:
        return self._get_user_id()

    # TODO remove once Microsoft support confirm their SSL certificates are always valid...
    def _get_user_id(self, number_of_retries=10):
        """"""
        try:
            return self._service.GetUser().User.Id
        except URLError as error:
            if isinstance(error.reason, ssl.SSLError):
                self.logger.warning("SSL certificate error, retrying...")
                if number_of_retries > 0:
                    time.sleep(1)
                    return self._get_user_id(number_of_retries - 1)
                else:
                    raise error

    def next_page_token(self, response: sudsobject.Object, **kwargs: Mapping[str, Any]) -> Optional[Mapping[str, Any]]:
        """
        Default method for streams that don't support pagination
        """
        return None

    def send_request(self, params: Mapping[str, Any], customer_id: str, account_id: str = None) -> Mapping[str, Any]:
        request_kwargs = {
            "service_name": self.service_name,
            "customer_id": customer_id,
            "account_id": account_id,
            "operation_name": self.operation_name,
            "params": params,
        }
        request = self.client.request(**request_kwargs)
        return request

    def read_records(
        self,
        sync_mode: SyncMode,
        stream_slice: Mapping[str, Any] = None,
        stream_state: Mapping[str, Any] = None,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Mapping[str, Any]]:
        stream_state = stream_state or {}
        next_page_token = None
        account_id = str(stream_slice.get("account_id")) if stream_slice else None
        customer_id = str(stream_slice.get("customer_id")) if stream_slice else None

        while True:
            params = self.request_params(
                stream_state=stream_state,
                stream_slice=stream_slice,
                next_page_token=next_page_token,
                account_id=account_id,
            )
            response = self.send_request(params, customer_id=customer_id, account_id=account_id)
            for record in self.parse_response(response):
                yield self.transform(record, stream_slice)

            next_page_token = self.next_page_token(response, current_page_token=next_page_token)
            if not next_page_token:
                break

    def parse_response(self, response: sudsobject.Object, **kwargs) -> Iterable[Mapping]:
        if response is not None and hasattr(response, self.data_field):
            yield from self.client.asdict(response)[self.data_field]


class BingAdsCampaignManagementStream(BingAdsStream, ABC):
    service_name: str = "CampaignManagement"

    @property
    @abstractmethod
    def data_field(self) -> str:
        """
        Specifies root object name in a stream response
        """

    @property
    @abstractmethod
    def additional_fields(self) -> Optional[str]:
        """
        Specifies which additional fields to fetch for a current stream.
        Expected format: field names separated by space
        """

    def parse_response(self, response: sudsobject.Object, **kwargs) -> Iterable[Mapping]:
        if response is not None and hasattr(response, self.data_field):
            yield from self.client.asdict(response)[self.data_field]


class Accounts(BingAdsStream):
    """
    Searches for accounts that the current authenticated user can access.
    API doc: https://docs.microsoft.com/en-us/advertising/customer-management-service/searchaccounts?view=bingads-13
    Account schema: https://docs.microsoft.com/en-us/advertising/customer-management-service/advertiseraccount?view=bingads-13
    Stream caches incoming responses to be able to reuse this data in Campaigns stream
    """

    primary_key = "Id"
    # Stream caches incoming responses to avoid duplicated http requests
    use_cache: bool = True
    data_field: str = "AdvertiserAccount"
    service_name: str = "CustomerManagementService"
    operation_name: str = "SearchAccounts"
    additional_fields: str = "TaxCertificate AccountMode"
    # maximum page size
    page_size_limit: int = 1000

    def __init__(self, client: Client, config: Mapping[str, Any]) -> None:
        super().__init__(client, config)
        self._account_names = config.get("account_names", [])
        self._unique_account_ids = set()

    def next_page_token(self, response: sudsobject.Object, current_page_token: Optional[int]) -> Optional[Mapping[str, Any]]:
        current_page_token = current_page_token or 0
        if response is not None and hasattr(response, self.data_field):
            return None if self.page_size_limit > len(response[self.data_field]) else current_page_token + 1
        else:
            return None

    def stream_slices(
        self,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        user_id_predicate = {
            "Field": "UserId",
            "Operator": "Equals",
            "Value": self._user_id,
        }
        if self._account_names:
            for account_config in self._account_names:
                account_name_predicate = {"Field": "AccountName", "Operator": account_config["operator"], "Value": account_config["name"]}

                yield {"predicates": {"Predicate": [user_id_predicate, account_name_predicate]}}
        else:
            yield {"predicates": {"Predicate": [user_id_predicate]}}

    def request_params(
        self,
        next_page_token: Mapping[str, Any] = None,
        stream_slice: Mapping[str, Any] = None,
        **kwargs: Mapping[str, Any],
    ) -> MutableMapping[str, Any]:
        paging = self._service.factory.create("ns5:Paging")
        paging.Index = next_page_token or 0
        paging.Size = self.page_size_limit
        return {
            "PageInfo": paging,
            "Predicates": stream_slice["predicates"],
            "ReturnAdditionalFields": self.additional_fields,
        }

    def _transform_tax_fields(self, record: Mapping[str, Any]) -> Mapping[str, Any]:
        tax_certificates = record["TaxCertificate"].get("TaxCertificates", {}) if record.get("TaxCertificate") is not None else {}
        if tax_certificates and not isinstance(tax_certificates, list):
            tax_certificate_pairs = tax_certificates.get("KeyValuePairOfstringbase64Binary")
            if tax_certificate_pairs:
                record["TaxCertificate"]["TaxCertificates"] = tax_certificate_pairs
        return record

    def parse_response(self, response: sudsobject.Object, **kwargs) -> Iterable[Mapping]:
        if response is not None and hasattr(response, self.data_field):
            records = self.client.asdict(response)[self.data_field]
            for record in records:
                if record["Id"] not in self._unique_account_ids:
                    self._unique_account_ids.add(record["Id"])
                    yield self._transform_tax_fields(record)


class Campaigns(BingAdsCampaignManagementStream):
    """
    Gets the campaigns for all provided accounts.
    API doc: https://docs.microsoft.com/en-us/advertising/campaign-management-service/getcampaignsbyaccountid?view=bingads-13
    Campaign schema: https://docs.microsoft.com/en-us/advertising/campaign-management-service/campaign?view=bingads-13
    Stream caches incoming responses to be able to reuse this data in AdGroups stream
    """

    primary_key = "Id"
    # Stream caches incoming responses to avoid duplicated http requests
    use_cache: bool = True
    data_field: str = "Campaign"
    operation_name: str = "GetCampaignsByAccountId"
    additional_fields: Iterable[str] = [
        "AdScheduleUseSearcherTimeZone",
        "BidStrategyId",
        "CpvCpmBiddingScheme",
        "DynamicDescriptionSetting",
        "DynamicFeedSetting",
        "MaxConversionValueBiddingScheme",
        "MultimediaAdsBidAdjustment",
        "TargetImpressionShareBiddingScheme",
        "TargetSetting",
        "VerifiedTrackingSetting",
    ]
    campaign_types: Iterable[str] = ["Audience", "DynamicSearchAds", "Search", "Shopping", "PerformanceMax"]

    parent_key_to_foreign_key_map = {
        "AccountId": "account_id",
        "CustomerId": "customer_id",
    }

    def request_params(
        self,
        stream_slice: Mapping[str, Any] = None,
        **kwargs: Mapping[str, Any],
    ) -> MutableMapping[str, Any]:
        return {
            "AccountId": stream_slice["account_id"],
            "CampaignType": " ".join(self.campaign_types),
            "ReturnAdditionalFields": " ".join(self.additional_fields),
        }

    def stream_slices(
        self,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        accounts = Accounts(self.client, self.config)
        for _slice in accounts.stream_slices():
            for account in accounts.read_records(SyncMode.full_refresh, _slice):
                yield {"account_id": account["Id"], "customer_id": account["ParentCustomerId"]}


class AdGroups(BingAdsCampaignManagementStream):
    """
    Gets the ad groups for all provided accounts.
    API doc: https://docs.microsoft.com/en-us/advertising/campaign-management-service/getadgroupsbycampaignid?view=bingads-13
    AdGroup schema: https://docs.microsoft.com/en-us/advertising/campaign-management-service/adgroup?view=bingads-13
    Stream caches incoming responses to be able to reuse this data in Ads stream
    """

    primary_key = "Id"
    # Stream caches incoming responses to avoid duplicated http requests
    use_cache: bool = True
    data_field: str = "AdGroup"
    operation_name: str = "GetAdGroupsByCampaignId"
    additional_fields: str = "AdGroupType AdScheduleUseSearcherTimeZone CpmBid CpvBid MultimediaAdsBidAdjustment"

    parent_key_to_foreign_key_map = {"CampaignId": "campaign_id", "AccountId": "account_id", "CustomerId": "customer_id"}

    def request_params(
        self,
        stream_slice: Mapping[str, Any] = None,
        **kwargs: Mapping[str, Any],
    ) -> MutableMapping[str, Any]:
        return {"CampaignId": stream_slice["campaign_id"], "ReturnAdditionalFields": self.additional_fields}

    def stream_slices(
        self,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        campaigns = Campaigns(self.client, self.config)
        accounts = Accounts(self.client, self.config)
        for _slice in accounts.stream_slices():
            for account in accounts.read_records(SyncMode.full_refresh, _slice):
                for campaign in campaigns.read_records(
                    sync_mode=SyncMode.full_refresh, stream_slice={"account_id": account["Id"], "customer_id": account["ParentCustomerId"]}
                ):
                    yield {"campaign_id": campaign["Id"], "account_id": account["Id"], "customer_id": account["ParentCustomerId"]}


class Ads(BingAdsCampaignManagementStream):
    """
    Retrieves the ads for all provided accounts.
    API doc: https://docs.microsoft.com/en-us/advertising/campaign-management-service/getadsbyadgroupid?view=bingads-13
    Ad schema: https://docs.microsoft.com/en-us/advertising/campaign-management-service/ad?view=bingads-13
    """

    primary_key = "Id"
    data_field: str = "Ad"
    operation_name: str = "GetAdsByAdGroupId"
    additional_fields: str = "ImpressionTrackingUrls Videos LongHeadlines"
    ad_types: Iterable[str] = [
        "Text",
        "Image",
        "Product",
        "AppInstall",
        "ExpandedText",
        "DynamicSearch",
        "ResponsiveAd",
        "ResponsiveSearch",
    ]

    parent_key_to_foreign_key_map = {"AdGroupId": "ad_group_id", "AccountId": "account_id", "CustomerId": "customer_id"}

    def request_params(
        self,
        stream_slice: Mapping[str, Any] = None,
        **kwargs: Mapping[str, Any],
    ) -> MutableMapping[str, Any]:
        return {
            "AdGroupId": stream_slice["ad_group_id"],
            "AdTypes": {"AdType": self.ad_types},
            "ReturnAdditionalFields": self.additional_fields,
        }

    def stream_slices(
        self,
        **kwargs: Mapping[str, Any],
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        ad_groups = AdGroups(self.client, self.config)
        for slice in ad_groups.stream_slices(sync_mode=SyncMode.full_refresh):
            for ad_group in ad_groups.read_records(sync_mode=SyncMode.full_refresh, stream_slice=slice):
                yield {"ad_group_id": ad_group["Id"], "account_id": slice["account_id"], "customer_id": slice["customer_id"]}



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
import os
import socket
import ssl
import sys
import uuid
from datetime import datetime, timedelta, timezone
from functools import lru_cache
from typing import Any, Iterator, List, Mapping, Optional, Union
from urllib.error import URLError

import backoff
import pendulum
from airbyte_cdk.models import FailureType
from airbyte_cdk.utils import AirbyteTracedException
from bingads.authorization import AuthorizationData, OAuthTokens, OAuthWebAuthCodeGrant
from bingads.exceptions import OAuthTokenRequestException
from bingads.service_client import ServiceClient
from bingads.util import errorcode_of_exception
from bingads.v13.bulk import BulkServiceManager, DownloadParameters
from bingads.v13.reporting.exceptions import ReportingDownloadException
from bingads.v13.reporting.reporting_service_manager import ReportingServiceManager
from suds import WebFault, sudsobject

FILE_TYPE = "Csv"
TIMEOUT_IN_MILLISECONDS = 3_600_000


class Client:
    api_version: int = 13
    refresh_token_safe_delta: int = 10  # in seconds
    logger: logging.Logger = logging.getLogger("airbyte")
    # retry on: rate limit errors, auth token expiration, internal errors
    # https://docs.microsoft.com/en-us/advertising/guides/services-protocol?view=bingads-13#throttling
    # https://docs.microsoft.com/en-us/advertising/guides/operation-error-codes?view=bingads-13
    retry_on_codes: Iterator[str] = ["117", "207", "4204", "109", "0"]
    max_retries: int = 5
    # A backoff factor to apply between attempts after the second try
    # {retry_factor} * (2 ** ({number of total retries} - 1))
    retry_factor: int = 15
    # environments supported by Microsoft Advertising: sandbox, production
    environment: str = "production"
    # The time interval in milliseconds between two status polling attempts.
    report_poll_interval: int = 15000
    # Timeout of downloading report
    _download_timeout = 300000
    _max_download_timeout = 600000

    reports_start_date = None

    def __init__(
        self,
        tenant_id: str,
        reports_start_date: str = None,
        developer_token: str = None,
        client_id: str = None,
        client_secret: str = None,
        refresh_token: str = None,
        **kwargs: Mapping[str, Any],
    ) -> None:
        self.refresh_token = refresh_token
        self.developer_token = developer_token

        self.client_id = client_id
        self.client_secret = client_secret

        self.authentication = self._get_auth_client(client_id, tenant_id, client_secret)
        self.oauth: OAuthTokens = self._get_access_token()
        if reports_start_date:
            self.reports_start_date = pendulum.parse(reports_start_date).astimezone(tz=timezone.utc)

    def _get_auth_client(self, client_id: str, tenant_id: str, client_secret: str = None) -> OAuthWebAuthCodeGrant:
        # https://github.com/BingAds/BingAds-Python-SDK/blob/e7b5a618e87a43d0a5e2c79d9aa4626e208797bd/bingads/authorization.py#L390
        auth_creds = {
            "client_id": client_id,
            "redirection_uri": "",  # should be empty string
            "client_secret": None,
            "tenant": tenant_id,
        }
        # the `client_secret` should be provided for `non-public clients` only
        # https://docs.microsoft.com/en-us/advertising/guides/authentication-oauth-get-tokens?view=bingads-13#request-accesstoken
        if client_secret and client_secret != "":
            auth_creds["client_secret"] = client_secret
        return OAuthWebAuthCodeGrant(**auth_creds)

    @lru_cache(maxsize=4)
    def _get_auth_data(self, customer_id: str = None, account_id: Optional[str] = None) -> AuthorizationData:
        return AuthorizationData(
            account_id=account_id,
            customer_id=customer_id,
            developer_token=self.developer_token,
            authentication=self.authentication,
        )

    def _get_access_token(self) -> OAuthTokens:
        self.logger.info("Fetching access token ...")
        # clear caches to be able to use new access token
        self.get_service.cache_clear()
        self._get_auth_data.cache_clear()
        try:
            tokens = self.authentication.request_oauth_tokens_by_refresh_token(self.refresh_token)
        except OAuthTokenRequestException as e:
            raise AirbyteTracedException(
                message=str(e),
                internal_message="Failed to get OAuth access token by refresh token. "
                "The user could not be authenticated as the grant is expired. The user must sign in again.",
                failure_type=FailureType.config_error,
            )
        return tokens

    def is_token_expiring(self) -> bool:
        """
        Performs check if access token expiring in less than refresh_token_safe_delta seconds
        """
        token_total_lifetime: timedelta = datetime.utcnow() - self.oauth.access_token_received_datetime
        token_updated_expires_in: int = self.oauth.access_token_expires_in_seconds - token_total_lifetime.seconds
        return False if token_updated_expires_in > self.refresh_token_safe_delta else True

    def should_give_up(self, error: Union[WebFault, URLError, ReportingDownloadException]) -> bool:
        if isinstance(error, URLError):
            if (
                isinstance(error.reason, socket.timeout)
                or isinstance(error.reason, ssl.SSLError)
                or isinstance(error.reason, socket.gaierror)  # temporary failure in name resolution
            ):
                return False
        if isinstance(error, ReportingDownloadException):
            self.logger.info("Reporting file download tracking status timeout.")
            if self._download_timeout < self._max_download_timeout:
                self._download_timeout = self._download_timeout + 10000
                self.logger.info(f"Increasing time of timeout to {self._download_timeout}")
            return False

        error_code = str(errorcode_of_exception(error))
        give_up = error_code not in self.retry_on_codes
        if give_up:
            self.logger.error(f"Giving up for returned error code: {error_code}. Error details: {self._get_error_message(error)}")
        return give_up

    def _get_error_message(self, error: WebFault) -> str:
        return str(self.asdict(error.fault)) if hasattr(error, "fault") else str(error)

    def log_retry_attempt(self, details: Mapping[str, Any]) -> None:
        _, exc, _ = sys.exc_info()
        self.logger.info(
            f"Caught retryable error: {self._get_error_message(exc)} after {details['tries']} tries. Waiting {details['wait']} seconds then retrying..."
        )

    def request(self, **kwargs: Mapping[str, Any]) -> Mapping[str, Any]:
        return backoff.on_exception(
            backoff.expo,
            (WebFault, URLError, ReportingDownloadException),
            max_tries=self.max_retries,
            factor=self.retry_factor,
            jitter=None,
            on_backoff=self.log_retry_attempt,
            giveup=self.should_give_up,
        )(self._request)(**kwargs)

    def _request(
        self,
        service_name: Optional[str],
        operation_name: str,
        customer_id: Optional[str],
        account_id: Optional[str],
        params: Mapping[str, Any],
        is_report_service: bool = False,
    ) -> Mapping[str, Any]:
        """
        Executes appropriate Service Operation on Bing Ads API
        """
        if self.is_token_expiring():
            self.oauth = self._get_access_token()

        if is_report_service:
            service = self._get_reporting_service(customer_id=customer_id, account_id=account_id)
        else:
            service = self.get_service(service_name=service_name, customer_id=customer_id, account_id=account_id)
        if operation_name == "download_report":
            params["download_parameters"].timeout_in_milliseconds = self._download_timeout
        return getattr(service, operation_name)(**params)

    @lru_cache(maxsize=4)
    def get_service(
        self,
        service_name: str,
        customer_id: str = None,
        account_id: Optional[str] = None,
    ) -> ServiceClient:
        return ServiceClient(
            service=service_name,
            version=self.api_version,
            authorization_data=self._get_auth_data(customer_id, account_id),
            environment=self.environment,
        )

    @lru_cache(maxsize=4)
    def _get_reporting_service(
        self,
        customer_id: Optional[str] = None,
        account_id: Optional[str] = None,
    ) -> ServiceClient:
        return ReportingServiceManager(
            authorization_data=self._get_auth_data(customer_id, account_id),
            poll_interval_in_milliseconds=self.report_poll_interval,
            environment=self.environment,
        )

    @classmethod
    def asdict(cls, suds_object: sudsobject.Object) -> Mapping[str, Any]:
        """
        Converts nested Suds Object into serializable format.
        Input sample:
        {
            obj[] =
                {
                    value = 1
                },
                {
                    value = "str"
                },
        }
        Output sample: =>
        {'obj': [{'value': 1}, {'value': 'str'}]}
        """
        result: Mapping[str, Any] = {}

        for field, val in sudsobject.asdict(suds_object).items():
            if hasattr(val, "__keylist__"):
                result[field] = cls.asdict(val)
            elif isinstance(val, list):
                result[field] = []
                for item in val:
                    if hasattr(item, "__keylist__"):
                        result[field].append(cls.asdict(item))
                    else:
                        result[field].append(item)
            elif isinstance(val, datetime):
                result[field] = val.isoformat()
            else:
                result[field] = val
        return result

    def _bulk_service_manager(self, customer_id: Optional[str] = None, account_id: Optional[str] = None):
        return BulkServiceManager(
            authorization_data=self._get_auth_data(customer_id, account_id),
            poll_interval_in_milliseconds=5000,
            environment=self.environment,
        )

    def get_bulk_entity(
        self,
        download_entities: List[str],
        data_scope: List[str],
        customer_id: Optional[str] = None,
        account_id: Optional[str] = None,
        start_date: Optional[str] = None,
    ) -> str:
        """
        Return path with zipped csv archive
        """
        download_parameters = DownloadParameters(
            # campaign_ids=None,
            data_scope=data_scope,
            download_entities=download_entities,
            file_type=FILE_TYPE,
            last_sync_time_in_utc=start_date,
            result_file_directory=os.getcwd(),
            result_file_name=str(uuid.uuid4()),
            overwrite_result_file=True,  # Set this value true if you want to overwrite the same file.
            timeout_in_milliseconds=TIMEOUT_IN_MILLISECONDS,  # You may optionally cancel the download after a specified time interval.
        )
        bulk_service_manager = self._bulk_service_manager(customer_id=customer_id, account_id=account_id)
        return bulk_service_manager.download_file(download_parameters)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from datetime import datetime, timezone


def transform_bulk_datetime_format_to_rfc_3339(original_value: str) -> str:
    """
    Bing Ads Bulk API provides datetime fields in custom format with milliseconds: "04/27/2023 18:00:14.970"
    Return datetime in RFC3339 format: "2023-04-27T18:00:14.970+00:00"
    """
    return datetime.strptime(original_value, "%m/%d/%Y %H:%M:%S.%f").replace(tzinfo=timezone.utc).isoformat(timespec="milliseconds")


def transform_date_format_to_rfc_3339(original_value: str) -> str:
    """
    Bing Ads API provides date fields in custom format: "04/27/2023"
    Return date in RFC3339 format: "2023-04-27"
    """
    return datetime.strptime(original_value, "%m/%d/%Y").replace(tzinfo=timezone.utc).strftime("%Y-%m-%d")


def transform_report_hourly_datetime_format_to_rfc_3339(original_value: str) -> str:
    """
    Bing Ads API reports with hourly aggregation provides date fields in custom format: "2023-11-04|11"
    Return date in RFC3339 format: "2023-11-04T11:00:00+00:00"
    """
    return datetime.strptime(original_value, "%Y-%m-%d|%H").replace(tzinfo=timezone.utc).isoformat(timespec="seconds")
